<?php

namespace Drupal\cforge\Installer;

use Drupal\user\Entity\User;
use Drupal\node\Entity\Node;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Extension\ExtensionDiscovery;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Installer\Exception\InstallerException;

/**
 * Choose between different Cforge flavours.
 */
class ChooseFlavour extends FormBase {

  private $installer;
  private $accountSwitcher;

  /**
   * Constructor.
   */
  public function __construct($module_installer, $account_switcher) {
    $this->installer = $module_installer;
    $this->accountSwitcher = $account_switcher;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('module_installer'),
      $container->get('account_switcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'cforge_flavour_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['#title'] = t('Choose flavour');
    foreach (CFORGE_FLAVOURS as $mod => $name) {
      module_set_weight($mod, 100);
    }
    $form['flavour'] = [
      '#title' => $this->t('Flavour', [], ['context' => 'Installation variety']),
      '#type' => 'radios',
      '#options' => CFORGE_FLAVOURS,
      '#default_value' => 'hamlets',
      '#required' => TRUE,
      '#weight' => 4,
    ];
    foreach ($form['flavour']['#options'] as $module => $name) {
      if ($message = static::installFlavour($module, TRUE)) {
        drupal_set_message($message, 'warning');
      }
    }

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => 'submit',
      '#weight' => 10,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $flavour_module = $form_state->getValue('flavour');

    if (!$flavour_module) {
      throw new InstallerException('Flavour not specified');
    }
    \Drupal::configFactory()->getEditable('cforge.settings')
      ->set('flavour', $flavour_module)
      ->save();
    $this->install($flavour_module);
  }


  function install($flavour) {
    module_load_install('cforge');
    drupal_set_message('Clear config before installing '.$flavour);
    cforge_clear_preexisting_config($flavour);

    if ($error = $this->installFlavour($flavour, TRUE)) {
      throw new InstallerException($error);
    }
    // Can't remember why we switch user
    $this->accountSwitcher->switchTo(User::load(1));
    if ($message = $this->installFlavour($flavour, FALSE)) {
      drupal_set_message($message, 'warning');
    }
    $this->accountSwitcher->switchBack();

    // cf_ladmin if it is available.
    $discovery = new ExtensionDiscovery(\Drupal::root());
    $available = $discovery->scan('module');
  }


  /**
   * Handle the installation of a flavour and its default content.
   *
   * @param string $module
   *   The name of the module to install.
   * @param bool $test
   *   Don't install, just check if directories are writable.
   *
   * @note this requires that drupal has permission to write inside the /profiles/directory
   * @note Export default_content with drush dcer node 1 --folder
   */
  function installFlavour($module, $test) {
    // The default_content module doesn't handle multiple languages.
    $content_path = DRUPAL_ROOT . '/' . drupal_get_path('module', $module) . '/content';
    $curr_lang_id = \Drupal::languageManager()->getDefaultLanguage()->getId();
    @unlink($content_path);
    symlink($content_path . '-' . $curr_lang_id, $content_path);
    if (file_exists($content_path)) {
      if (!$test) {
        $this->installer->install(['default_content', $module]);
        $this->installer->uninstall(['default_content', 'hal', 'serialization']);
        // Assume all the default content is public
        // set imported nodes to public using the cforge flag for that purpose.
        // set node owners to 1 because AccountSwitcher doesn't work.
        foreach (Node::loadMultiple() as $node) {
          $node->setOwnerId(1);
          if ($node->bundle() == 'page') {
            cforge_node_set_anonaccess($node->id(), TRUE);
          }
          // To get around a deficiency in default_content which fails to export Path
          // aliases we're going to set drupal's homepage using the hacked uuid.
          // @see https://www.drupal.org/node/2649646
          if (substr($node->uuid(), 0, 8) == '00000000') {
            $config = \Drupal::configFactory()->getEditable('system.site');
            $temp = $config->get('page');
            // The crux of the matter.
            $temp['front'] = '/node/' . $node->id();
            $config->set('page', $temp)->save();
          }
          $node->save();
        }
      }
    }
    else {
      return t(
        "Unable to install @module because @dir is not writable",
        ['@module' => $module, '@dir' => $content_path]
      );
    }
    @unlink($content_path);

  }

}
